#include "lusol.h"
/*--------------------------------------------*/
int ilu0(struct csr_t *csr, struct lu_t *lu)
/*
RETURN
0 : normal return
k : encountered a zero pivot at step k
*/
{
  int j,k,jj,jrow,jw;

  int n = csr->n;
  int nnz = csr->nnz;
  int *ia = csr->ia;
  int *ja = csr->ja;
  REAL *a0 = csr->a;

  // float entries of ILU0 precon
  REAL *a = (REAL *) malloc(nnz*sizeof(REAL));
  memcpy(a, a0, nnz*sizeof(REAL));

  // pointer of upper part of each row
  int *uptr = (int *) malloc(n*sizeof(int));

  /* initialize iw to 0 in parallel*/
  int *iw = (int *) calloc(n, sizeof(int));
 
  int nnzl = 0;
  /* main loop */
  for (k=0; k<n; k++)
  {
    /* working on row k */
    int j1 = ia[k];
    int j2 = ia[k+1];

    /* mark nnz in iw */    
    for (j=j1; j<j2; j++)
      iw[ja[j-1]-1] = j;

    jrow = -1;
    for (j=j1; j<j2; j++)
    {
      jrow = ja[j-1];

      if (jrow < k+1)
      {
        REAL tl = a[j-1]*a[uptr[jrow-1]-1];

        a[j-1] = tl;

        // linear combination
        for (jj=uptr[jrow-1]+1; jj<ia[jrow]; jj++)
          if (jw = iw[ja[jj-1]-1])
            a[jw-1] -= tl*a[jj-1];
      }
      else
        break;
    }

    uptr[k] = j;

    if (jrow != k+1 || a[j-1] == 0.0)	
      return (k);
    
    a[j-1] = 1.0 / a[j-1];

    /* reset iw */
    for (j=j1; j<j2; j++)
      iw[ja[j-1]-1] = 0;      
  }

  // nnz in l part
  for (k=0; k<n; k++)
    nnzl += (uptr[k] - ia[k]);

  // split precon to two CSRs, l and u 
  lu->l.n = n;
  lu->l.nnz = nnzl;
  lu->l.ia = (int *) malloc((n+1)*sizeof(int));
  lu->l.ja = (int *) malloc(nnzl*sizeof(int));
  lu->l.a = (REAL *) malloc(nnzl*sizeof(REAL));

  lu->u.n = n;
  lu->u.nnz = nnz - nnzl;

  lu->u.ia = (int *) malloc((n+1)*sizeof(int));
  lu->u.ja = (int *) malloc(lu->u.nnz*sizeof(int));
  lu->u.a = (REAL *) malloc(lu->u.nnz*sizeof(REAL));

  int i1 = 0;
  int i2 = 0;
  lu->l.ia[0] = lu->u.ia[0] = 1;
  for (k=0; k<n; k++)
  {
    // l part
    for (j=ia[k]; j<uptr[k]; j++)
    {
      lu->l.ja[i1] = ja[j-1];
      lu->l.a[i1] = a[j-1];
      i1++;
    }
    lu->l.ia[k+1] = lu->l.ia[k] + uptr[k] - ia[k];

    // u part
    for (j=uptr[k]; j<ia[k+1]; j++)
    {
      lu->u.ja[i2] = ja[j-1];
      lu->u.a[i2] = a[j-1];
      i2++;
    }
    lu->u.ia[k+1] = lu->u.ia[k] + ia[k+1] - uptr[k];
  }

  free(a);
  free(uptr);
  free(iw);

/*
  for (j=0; j<=n; j++) printf("%d ", lu->l.ia[j]); printf("\n");
  for (j=0; j<lu->l.nnz; j++) printf("%d ", lu->l.ja[j]); printf("\n");
  for (j=0; j<lu->l.nnz; j++) printf("%f ", lu->l.a[j]); printf("\n\n");

  for (j=0; j<=n; j++) printf("%d ", lu->u.ia[j]); printf("\n");
  for (j=0; j<lu->u.nnz; j++) printf("%d ", lu->u.ja[j]); printf("\n");
  for (j=0; j<lu->u.nnz; j++) printf("%f ", lu->u.a[j]); printf("\n");
*/
  return (-1);
}

